/*
 * This file is part of the vpnd2 project
 *
 * (C) 2017 Andreas Steinmetz, ast@domdv.de
 * The contents of this file is licensed under the GPL version 2 or, at
 * your choice, any later version of this license.
 */

#include <sys/timerfd.h>
#include <poll.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <curl/curl.h>

#define BUFSIZE	65536
#define STRING	0
#define NUMBER	1
#define BOOLEAN	2

struct parsecfg
{
	const char const *name;
	const int const type;
	const int const min;
	const int const max;
	const int const offset;
};

struct config
{
	char *proxyhost;
	char *proxytype;
	char *capath;
	unsigned int proxyport;
	unsigned int conntimeout;
	unsigned int fast;
	unsigned int keepalive;
	unsigned int verifyhost;
	unsigned int verifychain;
	unsigned int vryproxyhost;
	unsigned int vryproxychain;
};

static const struct parsecfg const items[]=
{
	{"proxyhost",STRING,0,0,offsetof(struct config,proxyhost)},
	{"proxyport",NUMBER,1,65535,offsetof(struct config,proxyport)},
	{"proxytype",STRING,0,0,offsetof(struct config,proxytype)},
	{"connecttimeout",NUMBER,1,60,offsetof(struct config,conntimeout)},
	{"fastconnect",BOOLEAN,0,0,offsetof(struct config,fast)},
	{"keepalive",NUMBER,15,120,offsetof(struct config,keepalive)},
	{"verifyhost",BOOLEAN,0,0,offsetof(struct config,verifyhost)},
	{"verifychain",BOOLEAN,0,0,offsetof(struct config,verifychain)},
	{"capath",STRING,0,0,offsetof(struct config,capath)},
	{"verifyproxyhost",BOOLEAN,0,0,offsetof(struct config,vryproxyhost)},
	{"verifyproxychain",BOOLEAN,0,0,offsetof(struct config,vryproxychain)},
	{NULL,0,0,0,0}
};

static CURL *doconnect(char *tlshost,char *svrhost,int svrport,int efd,
	struct config *c)
{
	int pos;
	int len;
	int tfd;
	size_t l;
	curl_socket_t s;
	CURL *curl;
	struct itimerspec it;
	struct pollfd p[3];
	char url[270];
	char proxy[290];
	char bfr[2048];

	sprintf(url,"https://%s/",tlshost);
	sprintf(bfr,"CONNECT %s:%d HTTP/1.1\r\nHost: %s:%d\r\n\r\n",
		svrhost,svrport,svrhost,svrport);

	if(!(curl=curl_easy_init()))goto err1;
	if(curl_easy_setopt(curl,CURLOPT_CONNECT_ONLY,1L)!=CURLE_OK)goto err2;
	if(curl_easy_setopt(curl,CURLOPT_NOSIGNAL,1L)!=CURLE_OK)goto err2;
	if(c->conntimeout)if(curl_easy_setopt(curl,CURLOPT_CONNECTTIMEOUT,
		(long)c->conntimeout)!=CURLE_OK)goto err2;
	if(curl_easy_setopt(curl,CURLOPT_URL,url)!=CURLE_OK)goto err2;
	if(c->proxyhost)
	{
		sprintf(proxy,"%s://%s:%d",c->proxytype,c->proxyhost,
			c->proxyport);
		if(curl_easy_setopt(curl,CURLOPT_PROXY,proxy)!=CURLE_OK)
			goto err2;
		if(curl_easy_setopt(curl,CURLOPT_PROXY_SSL_VERIFYHOST,
			c->vryproxyhost?2L:0L)!=CURLE_OK)goto err2;
		if(curl_easy_setopt(curl,CURLOPT_PROXY_SSL_VERIFYPEER,
			c->vryproxychain?1L:0L)!=CURLE_OK)goto err2;
	}
	if(curl_easy_setopt(curl,CURLOPT_SSL_VERIFYHOST,c->verifyhost?2L:0L)
		!=CURLE_OK)goto err2;
	if(curl_easy_setopt(curl,CURLOPT_SSL_VERIFYPEER,c->verifychain?1L:0L)
		!=CURLE_OK)goto err2;
	if(curl_easy_setopt(curl,CURLOPT_SSLVERSION,
		CURL_SSLVERSION_TLSv1_2|CURL_SSLVERSION_MAX_DEFAULT)
		!=CURLE_OK)goto err2;
	if(c->capath)curl_easy_setopt(curl,CURLOPT_CAPATH,c->capath);
	if(c->keepalive)
	{
		if(curl_easy_setopt(curl,CURLOPT_TCP_KEEPINTVL,
			(long)c->keepalive)!=CURLE_OK)goto err2;
		if(curl_easy_setopt(curl,CURLOPT_TCP_KEEPIDLE,
			(long)c->keepalive)!=CURLE_OK)goto err2;
		if(curl_easy_setopt(curl,CURLOPT_TCP_KEEPALIVE,1L)!=CURLE_OK)
			goto err2;
	}
	else if(curl_easy_setopt(curl,CURLOPT_TCP_KEEPALIVE,0L)!=CURLE_OK)
		goto err2;
	if(curl_easy_setopt(curl,CURLOPT_TCP_FASTOPEN,c->fast?1L:0L)!=CURLE_OK)
		goto err2;
	if(curl_easy_setopt(curl,CURLOPT_TCP_NODELAY,1L)!=CURLE_OK)goto err2;
	if(curl_easy_setopt(curl,CURLOPT_NOPROGRESS,1L)!=CURLE_OK)goto err2;
	if((tfd=timerfd_create(CLOCK_MONOTONIC,TFD_NONBLOCK))==-1)goto err2;
	memset(&it,0,sizeof(it));
	it.it_value.tv_sec=c->conntimeout;
	if(timerfd_settime(tfd,0,&it,NULL))goto err3;
	if(curl_easy_perform(curl)!=CURLE_OK)goto err3;
	if(curl_easy_getinfo(curl,CURLINFO_ACTIVESOCKET,&s)!=CURLE_OK)goto err3;

	p[0].fd=efd;
	p[0].events=POLLIN;
	p[1].fd=tfd;
	p[1].events=POLLIN;
	p[2].fd=s;
	p[2].events=POLLOUT;

	for(len=strlen(bfr),pos=0;pos<len;pos+=l)
	{
		while(poll(p,3,-1)<1);
		if(p[0].revents&POLLIN)goto err3;
		if(p[1].revents&POLLIN)goto err3;
		if(!(p[2].revents&POLLOUT))goto err3;
		if(curl_easy_send(curl,bfr+pos,len-pos,&l)!=CURLE_OK)goto err3;
	}

	p[2].events=POLLIN;

	for(len=0;len<sizeof(bfr);)
	{
		while(poll(p,3,-1)<1);
		if(p[0].revents&POLLIN)goto err3;
		if(p[1].revents&POLLIN)goto err3;
		if(!(p[2].revents&POLLIN))goto err3;
		if(curl_easy_recv(curl,bfr+len,sizeof(bfr)-len,&l)!=CURLE_OK)
			goto err3;
		len+=l;
		for(pos=0,l=0;pos<len;pos++)
		{
			if(bfr[pos]=='\r')continue;
			else if(bfr[pos]=='\n')l++;
			else l=0;
			if(l==2)
			{
				if(len<12)goto err3;
				if(strncmp(bfr,"HTTP/1.0 200",12)&&
					strncmp(bfr,"HTTP/1.1 200",12))
						goto err3;
				close(tfd);
				return curl;
			}
		}
	}

err3:	close(tfd);
err2:	curl_easy_cleanup(curl);
err1:	return NULL;
}

static void tcpforward(int lcl,CURL *curl)
{
	curl_socket_t rmt;
	size_t len;
	int lrfill;
	int rlfill;
	int lev;
	int rev;
	struct pollfd p[2];
	char lclrmt[BUFSIZE];
	char rmtlcl[BUFSIZE];

	if(curl_easy_getinfo(curl,CURLINFO_ACTIVESOCKET,&rmt)!=CURLE_OK)
		return;

	p[0].fd=lcl;
	p[1].fd=rmt;

	lev=POLLIN;
	rev=POLLIN;
	lrfill=0;
	rlfill=0;

	while(lev||rev)
	{
		p[0].events=lev;
		p[1].events=rev;

		while(poll(p,2,-1)<1);

		if(p[0].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[0].revents&POLLIN)
		{
			if((len=read(lcl,lclrmt+lrfill,BUFSIZE-lrfill))<=0)
				break;
			else lrfill+=len;
		}

		if(p[1].revents&(POLLERR|POLLHUP|POLLNVAL))break;

		if(p[1].revents&POLLIN)
		{
			if(curl_easy_recv(curl,rmtlcl+rlfill,BUFSIZE-rlfill,
				&len)!=CURLE_OK)len=-1;
			if(len<=0)break;
			else rlfill+=len;
		}

		if(lrfill&&(p[1].revents&POLLOUT))
		{
			if(curl_easy_send(curl,lclrmt,lrfill,&len)==CURLE_OK)
			{
				lrfill-=len;
				if(len)memmove(lclrmt,lclrmt+len,lrfill);
			}
			else break;
		}

		if(rlfill&&(p[0].revents&POLLOUT))
		{
			if((len=write(lcl,rmtlcl,rlfill))>=0)
			{
				rlfill-=len;
				if(len)memmove(rmtlcl,rmtlcl+len,rlfill);
			}
			else break;
		}

		if(lrfill==BUFSIZE)lev=0;
		else lev=POLLIN;
		if(rlfill)lev|=POLLOUT;

		if(rlfill==BUFSIZE)rev=0;
		else rev=POLLIN;
		if(lrfill)rev|=POLLOUT;
	}
}

static int parseline(char *line,int num,void *base)
{       
	int i;
	unsigned long number;
	char *mem;
	char *item;
	char *value;
	char *ptr;
	
	item=strtok_r(line," \t\r\n",&mem);
	value=strtok_r(NULL," \t\r\n",&mem);
	
	if(!item||!value||!*item||!*value)goto err1;
	for(i=0;items[i].name;i++)if(!strcmp(item,items[i].name))break;
	if(!items[i].name)goto err1;
	
	switch(items[i].type)
	{
	case STRING:
		if(!*value||strlen(value)>255)goto err1;
		if(!(ptr=strdup(value)))goto err2;
		*((char **)(base+items[i].offset))=ptr;
		break;
	
	case NUMBER:
		if((number=strtoul(value,&mem,10))==ULONG_MAX||*mem)goto err1;
		if(number<items[i].min||number>items[i].max)goto err1;
		*((unsigned int *)(base+items[i].offset))=(unsigned int)number;
		break;

	case BOOLEAN:
		if(!strcmp(value,"enable"))
			*((unsigned int *)(base+items[i].offset))=1;
		else if(!strcmp(value,"disable"))
			*((unsigned int *)(base+items[i].offset))=0;
		else goto err1;
		break;

	default:goto err1;
	}

	return 0;

err2:   fprintf(stderr,"internal error\n");
	return -1;

err1:   fprintf(stderr,"configuration syntax error line %d\n",num);
	return -1;
}

static int parseconfig(char *fn,struct config *cfg)
{
	int num=0;
	FILE *fp;
	char *ptr;
	char line[8192];

        if(!(fp=fopen(fn,"r")))
        {
                fprintf(stderr,"can't open %s\n",fn);
                goto err1;
        }

        while(fgets(line,sizeof(line),fp))
        {
                num++;
                for(ptr=line;*ptr;ptr++)if(*ptr!=' '||*ptr!='\t')break;
                if(!*ptr||*ptr=='#'||*ptr=='\r'||*ptr=='\n')continue;
		if(parseline(ptr,num,cfg))goto err2;
	}

	fclose(fp);

	if(cfg->proxytype||cfg->proxyhost||cfg->proxyport)
	{
		if(!cfg->proxytype||!cfg->proxyhost||!cfg->proxyport)goto err3;
		if(strcmp(cfg->proxytype,"http")&&strcmp(cfg->proxytype,"https")
			&&strcmp(cfg->proxytype,"socks4")
			&&strcmp(cfg->proxytype,"socks4a")
			&&strcmp(cfg->proxytype,"socks5")
			&&strcmp(cfg->proxytype,"socks5h"))goto err3;
	}

	return 0;

err3:	fprintf(stderr,"configuration syntax error\n");
err2:	fclose(fp);
err1:	return -1;
}

int main(int argc,char *argv[])
{
	int port;
	CURL *c;
	char *host;
	struct config cfg;

	memset(&cfg,0,sizeof(cfg));
	close(1);

	if(fcntl(0,F_SETFL,O_RDWR|O_NONBLOCK))goto err1;

	if(!(host=getenv("VPN_PORT")))goto err1;
	port=atoi(host);
	if(port<1||port>65535)goto err1;
	if(!(host=getenv("VPN_HOST")))goto err1;
	if(!*host||strlen(host)>255)goto err1;
	if(argc==2)
	{
		if(parseconfig(argv[1],&cfg))goto err1;
	}
	else if(argc!=1)goto err1;
	if(!cfg.conntimeout)cfg.conntimeout=10;

	if(!(c=doconnect(host,"localhost",port,0,&cfg)))goto err1;
	if(write(0,"OK\n",3)!=3)goto err2;

	tcpforward(0,c);

	curl_easy_cleanup(c);
	close(0);
	return 0;

err2:	curl_easy_cleanup(c);
err1:	if(write(0,"NO\n",3)!=3);
	return 1;
}
